
// Range representation base class.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Pool' : Modified from Stroustrup.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#ifdef USE_POOL

class Pool {
  struct Link {Link* next ; } ;

  unsigned esize ;
  Link* head ;

  void grow() ;

  int nalloc ;  // debug
public :
  void init(unsigned n) ;

  void* alloc() ;
  void free(void* b) ;
} ;

inline void* Pool :: alloc() {

    printf("%d inline void* Pool :: alloc() in RangeRep.h\n", nalloc);

  if(head == 0) grow() ;
  Link* p = head ;
  head = p->next ;
  nalloc++ ;  // debug
  return p ;
}

inline void Pool :: free(void* b) {

    printf("%d inline void* Pool :: free() in RangeRep.h\n", nalloc);

  Link* p = (Link*) b ;
  p->next = head ;
  head = p ;
  nalloc-- ;  // debug
}


#define POOL_NEW void* operator new(size_t) { return pool.alloc() ; }

#define POOL_DELETE void operator delete(void* p) { pool.free(p) ; }

#define POOL_STATE static Pool pool ; friend void RangeInit() ;

#define MEMORY_POOL_DEFN(T) Pool T :: pool ;

#define MEMORY_POOL_DECS public : POOL_NEW POOL_DELETE protected : POOL_STATE

void RangeInit() ;

#endif // USE_POOL

enum Format {DIST_PRIMITIVE, DIST_COLLAPSED, DIST_BLOCK, DIST_CYCLIC,
             DIST_BLOCK_CYCLIC, DIST_STEP, DIST_IRREG} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Location': The local subscript.                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


struct Location {

  Dimension dim ;
  int crd ;

  int sub ;

  int blk ;
} ;


// Restriction operator on `Group'.

inline Group operator/(const Group& p, const Location& i) {
  Group result(p) ;
  result.restrict(i.dim, i.crd) ;
  return result ;
}

inline Group operator/(const Procs& p, const Location& i) {
  Group result(p) ;
  result.restrict(i.dim, i.crd) ;
  return result ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Block'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Block {
  int count ;

  int glb_bas ;
  int glb_stp ;

  int sub_bas ;
  int sub_stp ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RangeRep': Base class for representation of ranges.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class RangeRep {
public :
  // Constructors and destructors

  RangeRep(const int _extent,
           RangeRep* _kernel, const int _level, Dimension _dimension,
           const Format _distFormat) : 
      extent(_extent), kernel(_kernel), level(_level), dimension(_dimension),
      distFormat(_distFormat),
      refCount(0) {}

  virtual ~RangeRep() {}

  // Subranges.

  virtual inline RangeRep* subrng(const int _extent,
                                  const int _base, const int _stride) ;


  // Inquiries.

  int size() const {return extent ; }

  Dimension dim() const {return dimension ; }

  Format format() const {return distFormat ; }

  virtual RangeRep* tem() {return this ; }
  virtual int bas() const {return 0 ; }
  virtual int str() const {return 1 ; }

  virtual int loExtension() const {return 0 ; }
  virtual int hiExtension() const {return 0 ; }

  int lev() const {return level ; }
  virtual RangeRep* ker() const {return kernel ; }

  virtual RangeRep* shell() const {
    AdlibError("RangeRep :: shell: member undefined on this range") ;
    return 0 ;
  }

  virtual void location(Location* loc, const int glb) const = 0 ;

  virtual void block(Block*, const int) const {
    AdlibError("RangeRep :: block: member undefined on this range") ;
  }

  virtual void block2(Block*, const int,
                      const int, const int) const {
    AdlibError("RangeRep :: block2: member undefined on this range") ;
  }

  virtual void block3(Block*, const int,
                      const int, const int, const int) const {
    AdlibError("RangeRep :: block3: member undefined on this range") ;
  }

//  virtual int local(int*) const {
//    AdlibError("RangeRep :: local: member undefined on this range") ;
//    return 0 ;
//  }

  virtual void crds(int* ker_count, int* ker_bas, int* ker_str) const {
    *ker_count = extent ;
    *ker_bas   = 0 ;
    *ker_str   = 1 ;
  }

  virtual void crds2(int*, int*, int*,
                     const int, const int) const {
    AdlibError("RangeRep :: crds2: member undefined on this range") ;
  }

  virtual void crds3(int*, int*, int*,
                     const int, const int, const int) const {
    AdlibError("RangeRep :: crds3: member undefined on this range") ;
  }

  // Maps

  virtual int disp() const {
    return 0 ;
  }

  virtual int volume() const = 0 ;

  virtual int offset(const Location* i) const {
    return i->sub ;
  }

//  virtual int disp(const int sub) const {
//    return sub ;
//  }
//
//  virtual int step(const int sub_stp) const {
//    return sub_stp ;
//  }
//
//  virtual int disps(int* off_bas, int* off_stp,
//                    const int sub_bas, const int sub_stp) const {
//    *off_bas = sub_bas ;
//    *off_stp = sub_stp ;
//  }

  // Reference counting.

  void incr() {refCount++ ; }
  void decr() {if((--refCount) == 0) delete this ; }
  
  // Interim support for PCRC Fortran interface:

  virtual int glb(const int sub, const int) const {return sub ; }

  virtual int isAligned(RangeRep* x) const = 0 ;

  virtual void show() = 0 ;
protected :
  int extent ;
  RangeRep* kernel ;
  int level ;
  Dimension dimension ;
  const Format distFormat ;

private :
  int refCount ;

  // Interim support for PCRC Fortran interface:

  int g_lb, g_stride ;

friend class Range ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SubRangeRep': generic subrange.                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SubRangeRep : public RangeRep {
public :
  SubRangeRep(const int _extent, const int _base, const int _stride,
              RangeRep* _parent) :
          RangeRep(_extent, _parent->ker(), _parent->lev(),
                   _parent->dim(), _parent->format()),
          base(_base), stride(_stride), parent(_parent) {
    parent->incr() ;
  }

  ~SubRangeRep() {
    parent->decr() ;
  }

  RangeRep* tem() {return parent ; }

  int str() const {return stride ; }
  int bas() const {return base ; }

  RangeRep* subrng(const int _extent,
                   const int _base, const int _stride) {
    return new SubRangeRep(_extent,
                           base + stride * _base, stride * _stride,
                           parent) ;
  }

  RangeRep* shell() const {
    return parent->shell() ;
  }

  void location(Location* loc, const int glb) const {
    parent->location(loc, base + stride * glb) ;
  }

//  int disp(const int sub) const {
//    return parent->disp(sub) ;
//  }
//
//  int step(const int sub_stp) const {
//    return parent->step(sub_stp) ;
//  }

  int offset(const Location* i) const {
    return parent->offset(i) ;
  }

  void block(Block* blk, const int ker_glb) const {
    parent->block3(blk, ker_glb, extent, base, stride) ;
  }

  void block2(Block* blk, const int ker_glb,
              const int count, const int _base) const {
    parent->block3(blk, ker_glb, count, base + stride * _base, stride) ;
  }

  void block3(Block* blk, const int ker_glb,
              const int count, const int _base, const int _stride) const {
    parent->block3(blk, ker_glb,
                   count, base + stride * _base, stride * _stride) ;
  }

  void crds(int* ker_count, int* ker_bas, int* ker_str) const {
    parent->crds3(ker_count, ker_bas, ker_str, extent, base, stride) ;
  }

  void crds2(int* ker_count, int* ker_bas, int* ker_str,
             const int count, const int _base) const {
    parent->crds3(ker_count, ker_bas, ker_str,
                  count, base + stride * _base, stride) ;
  }

  void crds3(int* ker_count, int* ker_bas, int* ker_str,
             const int count, const int _base, const int _stride) const {
    parent->crds3(ker_count, ker_bas, ker_str,
                  count, base + stride * _base, stride * _stride) ;
  }

  int disp() const {
    return parent->disp() ;
  }

  int volume() const {
    return parent->volume() ;
  }

  int glb(const int sub, const int ker_glb) const {
    return (parent->glb(sub, ker_glb) - base) / stride ;
  }

  int isAligned(RangeRep* x) const {
    if(x == (RangeRep*) this) return 1 ;
    if(x->tem() == parent || parent->isAligned(x->tem()))
      return x->size() == extent && x->bas() == base && x->str() == stride ;
    else
      return 0 ;
  }

  void show() {
    AdlibPrintf("SubRangeRep :: show: "
                "size = %d bas = %d str = %d par = %p\n",
                 extent, base, stride, parent) ;
  }

protected :

  int base, stride ;
  RangeRep* parent ;
} ;

RangeRep* RangeRep :: subrng(const int _extent,
                             const int _base, const int _stride) {
  return new SubRangeRep(_extent, _base, _stride, this) ;
}


// ---------------------------------------------------
//
// Things to do:
//
//   Things that return `RangeRep*' should return `const RangeRep*'?
//       Doesn't really work, because often need to be able to update
//       the reference count.
//
//   Simplified sub-range classes for contiguous subranges.

